import timeit

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.linalg as LA

from animate_chain import animate_chain
from finite_difference import finite_difference
from hc_obj import hc_obj


# for people who cannot see the plots
if mpl.get_backend() == 'agg':
    mpl.use('WebAgg')
print(f"backend: {mpl.get_backend()}")


# a flag on whether draw animation for each iteration.
# plot_figures = False
plot_figures = True


# Number of mass points
N     = 40
tol   = 1E-3
maxit = 10000


# Chain end points
xi = [-2, 1]  # initial (y1, z1)
xf = [ 2, 1]  # final   (yN, zN)

# Definining constants in a dictionary
param = {
    'L': 1,
    'D': (70/40)*N,
    'm': 4/N,
    'g': 9.81,
    'N': N,
    'xi': xi,
    'xf': xf,
}

# Initial guess: linear interpolation between the tips
y = np.linspace(xi[0], xf[0], N+2)
z = np.linspace(xi[1], xf[1], N+2)

# Eliminate the tips from the decision variables and stack them together
x = np.column_stack((y[1:-1], z[1:-1])).reshape(2*N)

# Hessian approximation
# Bk = 10000*np.eye(np.size(x)) # this works also without line search
Bk =  np.eye(np.size(x))

# Get the objective value and the gradient at initial point
[F, J] = finite_difference(hc_obj, x, param)

# A check whether the function you wrote is  correct
if F > 46.93 or F < 46.92 or LA.norm(J) > 6.21 or LA.norm(J) < 6.20:
    raise ValueError(f"The outputs from your function finite_difference() or hc_obj() for the given initial point are not correct. You have F={F}, ||J||={LA.norm(J)}")


# Printing header: iterate number, gradient norm, objective value, step norm, stepsize
print("It. \t | ||grad_f||\t\t | f\t\t | ||dvar||\t | t  \n")

gradnorm_gradient = LA.norm(J)
time_ref = timeit.default_timer()
if plot_figures:
    x_traj = []
    dx_traj = []

# Iterative loop
for k in range(1, maxit+1):
    # Obtaining search direction
    dx = -LA.solve(Bk, J.T).flatten()

    # We check Armijo's condition to find a step length
    t     = 1.0    # Initial step length
    beta  = 0.8    # Shrinking factor
    gamma = 0.1    # Minimal decrease requirement

    x_new = x + t*dx  # Candidate for the next step

    # Iterate on Armijo's condition
    while np.any(hc_obj(x_new, param) > F + gamma*t*J@dx):
        t = beta * t
        x_new = x + t*dx

    # Assign the step
    [F_new, J_new] = finite_difference(hc_obj, x_new, param)

    # Updating variables
    x = x_new
    J = J_new
    F = F_new

    # Every 10 iterations print the header again
    if k % 10 == 0:
        print("\n")
        print("It. \t | ||grad_f||\t\t | f\t\t | ||dvar||\t | t  \n")

    # Print some useful information
    print(f"{k}\t | {LA.norm(J)}\t | {F}\t | {LA.norm(dx)}\t | {t} \n")

    if plot_figures:
        x_traj.append(x)
        dx_traj.append(dx)

    if LA.norm(J) < tol:
        print("Convergence achieved.")
        break


time_exec = timeit.default_timer() - time_ref
print(f"The execution time is {time_exec}s")


# plotting iterations
if plot_figures:
    ani = animate_chain(x_traj, dx_traj)

# Plotting optimal solution 
plt.figure(1)
y = x[0:2*N:2]
z = x[1:2*N:2]
plt.subplot(2,1,1)
plt.plot(y, z, 'b--')
plt.plot(y, z, 'ro', mfc='none')
plt.title(r'Position of chain at optimal solution')
plt.xlim([-2, 2])
plt.ylim([-3, 1])


plt.show()
